/* **********************************************************************

 * _apRateConvertPolynomialOrder3:  Rate convert block by 3rd order
 *                                  polynomial interpolation
 *                                  between four proximal samples
 *
 *      (x0,y0), (x1,y1) are two points preceding unknown point
 *      (x2,y2), (x3,y3) are two points following unknown point
 *      y =  -y0(x-x1)(x-x2)(x-x3)/6
 *          +y1(x-x0)(x-x2)(x-x3)/2
 *          -y2(x-x0)(x-x1)(x-x3)/2
 *          +y3(x-x0)(x-x1)(x-x2)/6
 *      
 *    =>     [y3(x-x0) - y0(x-x3)](x-x1)(x-x2)/6
 *          +[y1(x-x2) - y2(x-x1)](x-x0)(x-x3)/2
 * ********************************************************************** */

void
_apRateConvertPolynomialOrder3(float *in, float *out, RateConvertData * data)
{
    register float *src = in;
    register float *dst = out;
    register float *ptr;

    register double increment = data->fractionalPtrIncrement;
    register double x;		/* fractional counter */

    register float s;		/* one sample */

    register float a, b, c, d, e, f;	/* intermediate values */

    register int n;

    int length = data->atomicInBlockLength;
    int outLen = data->atomicOutBlockLength;
    
    register float len = (float) (length);

    register float one = 1.0;
    register float two = 2.0;
    register float three = 3.0;
    register float onehalf = 0.5;
    register float onesixth = 0.166666666667;

    register float *old = data->baseMemory;

    /* 
     * start off at the first sample in the source and dest. buffers
     */
    x = data->fractionalPtr;

    while ((b = x) < one && outLen--) {
	a = b + one;
	c = b - one;
	d = b - two;

	e = a * b;
	f = c * d;

	s =
	    onehalf * (a * f * old[1] - e * d * old[0]) +
	    onesixth * (e * c * src[0] - b * f * old[2]);

	*dst++ = s;

	x += increment;		/* push the fraction along */
    }

    while ((b = x - one) < one && outLen--) {
	a = b + one;
	c = b - one;
	d = b - two;

	e = a * b;
	f = c * d;

	s =
	    onehalf * (a * f * old[0] - e * d * src[0]) +
	    onesixth * (e * c * src[1] - b * f * old[1]);

	*dst++ = s;

	x += increment;		/* push the fraction along */
    }

    while ((b = x - two) < one && outLen--) {
	a = b + one;
	c = b - one;
	d = b - two;

	e = a * b;
	f = c * d;

	s =
	    onehalf * (a * f * src[0] - e * d * src[1]) +
	    onesixth * (e * c * src[2] - b * f * old[0]);

	*dst++ = s;

	x += increment;		/* push the fraction along */
    }

    n = (int) x;		/* sample number before fractional counter */
    if (n < 2)
        n = 2;
    while (x < len && outLen--) {
	ptr = src + n - 2;

	b = x - (float) n;
	a = b + one;
	c = b - one;
	d = b - two;

	e = a * b;
	f = c * d;

	s =
	    onehalf * (a * f * ptr[0] - e * d * ptr[1]) +
	    onesixth * (e * c * ptr[2] - b * f * ptr[-1]);

	*dst++ = s;

	x += increment;		/* push the fraction along */
        n = (int) x;
    }

    src = in + length;

    old[2] = src[-3];
    old[1] = src[-2];
    old[0] = src[-1];

    data->fractionalPtr = x - len;

    data->sampsOut = (dst - out);
}				/* ---- end _apRateConvertPolynomialOrder3() ---- */
